@@ -9,12 +9,14 @@ import atexit
 import os
 import sys
 import signal
+import traceback
+import logging
 
+from tornado import ioloop
 import zmq
-from zmq.eventloop import ioloop
+from zmq.eventloop import ioloop as zmq_ioloop
 from zmq.eventloop.zmqstream import ZMQStream
 
-from IPython.core.ultratb import FormattedTB
 from IPython.core.application import (
     BaseIPythonApplication, base_flags, base_aliases, catch_config_error
 )
@@ -25,7 +27,7 @@ from IPython.core.shellapp import (
 from IPython.utils import io
 from ipython_genutils.path import filefind, ensure_dir_exists
 from traitlets import (
-    Any, Instance, Dict, Unicode, Integer, Bool, DottedObjectName, Type,
+    Any, Instance, Dict, Unicode, Integer, Bool, DottedObjectName, Type, default
 )
 from ipython_genutils.importstring import import_item
 from jupyter_core.paths import jupyter_runtime_dir
@@ -33,6 +35,7 @@ from jupyter_client import write_connection_file
 from jupyter_client.connect import ConnectionFileMixin
 
 # local imports
+from .iostream import IOPubThread
 from .heartbeat import Heartbeat
 from .ipkernel import IPythonKernel
 from .parentpoller import ParentPollerUnix, ParentPollerWindows
@@ -100,24 +103,31 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
     flags = Dict(kernel_flags)
     classes = [IPythonKernel, ZMQInteractiveShell, ProfileDir, Session]
     # the kernel class, as an importstring
-    kernel_class = Type('ipykernel.ipkernel.IPythonKernel', config=True,
+    kernel_class = Type('ipykernel.ipkernel.IPythonKernel',
                         klass='ipykernel.kernelbase.Kernel',
     help="""The Kernel subclass to be used.
 
     This should allow easy re-use of the IPKernelApp entry point
     to configure and launch kernels other than IPython's own.
-    """)
+    """).tag(config=True)
     kernel = Any()
     poller = Any() # don't restrict this even though current pollers are all Threads
     heartbeat = Instance(Heartbeat, allow_none=True)
     ports = Dict()
 
+    subcommands = {
+        'install': (
+            'ipykernel.kernelspec.InstallIPythonKernelSpecApp',
+            'Install the IPython kernel'
+        ),
+    }
+
     # connection info:
     connection_dir = Unicode()
-    def _connection_dir_default(self):
-        d = jupyter_runtime_dir()
-        ensure_dir_exists(d, 0o700)
-        return d
+
+    @default('connection_dir')
+    def _default_connection_dir(self):
+        return jupyter_runtime_dir()
 
     @property
     def abs_connection_file(self):
@@ -126,35 +136,39 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         else:
             return self.connection_file
 
-
     # streams, etc.
-    no_stdout = Bool(False, config=True, help="redirect stdout to the null device")
-    no_stderr = Bool(False, config=True, help="redirect stderr to the null device")
+    no_stdout = Bool(False, help="redirect stdout to the null device").tag(config=True)
+    no_stderr = Bool(False, help="redirect stderr to the null device").tag(config=True)
     outstream_class = DottedObjectName('ipykernel.iostream.OutStream',
-        config=True, help="The importstring for the OutStream factory")
+        help="The importstring for the OutStream factory").tag(config=True)
     displayhook_class = DottedObjectName('ipykernel.displayhook.ZMQDisplayHook',
-        config=True, help="The importstring for the DisplayHook factory")
+        help="The importstring for the DisplayHook factory").tag(config=True)
 
     # polling
-    parent_handle = Integer(int(os.environ.get('JPY_PARENT_PID') or 0), config=True,
+    parent_handle = Integer(int(os.environ.get('JPY_PARENT_PID') or 0),
         help="""kill this process if its parent dies.  On Windows, the argument
         specifies the HANDLE of the parent process, otherwise it is simply boolean.
-        """)
-    interrupt = Integer(int(os.environ.get('JPY_INTERRUPT_EVENT') or 0), config=True,
+        """).tag(config=True)
+    interrupt = Integer(int(os.environ.get('JPY_INTERRUPT_EVENT') or 0),
         help="""ONLY USED ON WINDOWS
         Interrupt this process when the parent is signaled.
-        """)
+        """).tag(config=True)
 
     def init_crash_handler(self):
-        # Install minimal exception handling
-        sys.excepthook = FormattedTB(mode='Verbose', color_scheme='NoColor',
-                                     ostream=sys.__stdout__)
+        sys.excepthook = self.excepthook
+
+    def excepthook(self, etype, evalue, tb):
+        # write uncaught traceback to 'real' stderr, not zmq-forwarder
+        traceback.print_exception(etype, evalue, tb, file=sys.__stderr__)
 
     def init_poller(self):
         if sys.platform == 'win32':
             if self.interrupt or self.parent_handle:
                 self.poller = ParentPollerWindows(self.interrupt, self.parent_handle)
-        elif self.parent_handle:
+        elif self.parent_handle and self.parent_handle != 1:
+            # PID 1 (init) is special and will never go away,
+            # only be reassigned.
+            # Parent polling doesn't work if ppid == 1 to start with.
             self.poller = ParentPollerUnix()
 
     def _bind_socket(self, s, port):
@@ -201,7 +215,9 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
             self.connection_file = filefind(self.connection_file, ['.', self.connection_dir])
         except IOError:
             self.log.debug("Connection file not found: %s", self.connection_file)
-            # This means I own it, so I will clean it up:
+            # This means I own it, and I'll create it in this directory:
+            ensure_dir_exists(os.path.dirname(self.abs_connection_file), 0o700)
+            # Also, I will clean it up:
             atexit.register(self.cleanup_connection_file)
             return
         try:
@@ -222,11 +238,6 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         self.shell_port = self._bind_socket(self.shell_socket, self.shell_port)
         self.log.debug("shell ROUTER Channel on port: %i" % self.shell_port)
 
-        self.iopub_socket = context.socket(zmq.PUB)
-        self.iopub_socket.linger = 1000
-        self.iopub_port = self._bind_socket(self.iopub_socket, self.iopub_port)
-        self.log.debug("iopub PUB Channel on port: %i" % self.iopub_port)
-
         self.stdin_socket = context.socket(zmq.ROUTER)
         self.stdin_socket.linger = 1000
         self.stdin_port = self._bind_socket(self.stdin_socket, self.stdin_port)
@@ -237,6 +248,19 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         self.control_port = self._bind_socket(self.control_socket, self.control_port)
         self.log.debug("control ROUTER Channel on port: %i" % self.control_port)
 
+        self.init_iopub(context)
+
+    def init_iopub(self, context):
+        self.iopub_socket = context.socket(zmq.PUB)
+        self.iopub_socket.linger = 1000
+        self.iopub_port = self._bind_socket(self.iopub_socket, self.iopub_port)
+        self.log.debug("iopub PUB Channel on port: %i" % self.iopub_port)
+        self.configure_tornado_logger()
+        self.iopub_thread = IOPubThread(self.iopub_socket, pipe=True)
+        self.iopub_thread.start()
+        # backward-compat: wrap iopub socket API in background thread
+        self.iopub_socket = self.iopub_thread.background_socket
+
     def init_heartbeat(self):
         """start the heart beating"""
         # heartbeat doesn't share context, because it mustn't be blocked
@@ -267,7 +291,8 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         for line in lines:
             self.log.info(line)
         # also raw print to the terminal if no parent_handle (`ipython kernel`)
-        if not self.parent_handle:
+        # unless log-level is CRITICAL (--quiet)
+        if not self.parent_handle and self.log_level < logging.CRITICAL:
             io.rprint(_ctrl_c_message)
             for line in lines:
                 io.rprint(line)
@@ -289,11 +314,39 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         """Redirect input streams and set a display hook."""
         if self.outstream_class:
             outstream_factory = import_item(str(self.outstream_class))
-            sys.stdout = outstream_factory(self.session, self.iopub_socket, u'stdout')
-            sys.stderr = outstream_factory(self.session, self.iopub_socket, u'stderr')
+            sys.stdout = outstream_factory(self.session, self.iopub_thread, u'stdout')
+            sys.stderr = outstream_factory(self.session, self.iopub_thread, u'stderr')
         if self.displayhook_class:
             displayhook_factory = import_item(str(self.displayhook_class))
-            sys.displayhook = displayhook_factory(self.session, self.iopub_socket)
+            self.displayhook = displayhook_factory(self.session, self.iopub_socket)
+            sys.displayhook = self.displayhook
+
+        self.patch_io()
+
+    def patch_io(self):
+        """Patch important libraries that can't handle sys.stdout forwarding"""
+        try:
+            import faulthandler
+        except ImportError:
+            pass
+        else:
+            # Warning: this is a monkeypatch of `faulthandler.enable`, watch for possible
+            # updates to the upstream API and update accordingly (up-to-date as of Python 3.5):
+            # https://docs.python.org/3/library/faulthandler.html#faulthandler.enable
+
+            # change default file to __stderr__ from forwarded stderr
+            faulthandler_enable = faulthandler.enable
+            def enable(file=sys.__stderr__, all_threads=True, **kwargs):
+                return faulthandler_enable(file=file, all_threads=all_threads, **kwargs)
+
+            faulthandler.enable = enable
+
+            if hasattr(faulthandler, 'register'):
+                faulthandler_register = faulthandler.register
+                def register(signum, file=sys.__stderr__, all_threads=True, chain=False, **kwargs):
+                    return faulthandler_register(signum, file=file, all_threads=all_threads,
+                                                 chain=chain, **kwargs)
+                faulthandler.register = register
 
     def init_signal(self):
         signal.signal(signal.SIGINT, signal.SIG_IGN)
@@ -307,18 +360,31 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
 
         kernel = kernel_factory(parent=self, session=self.session,
                                 shell_streams=[shell_stream, control_stream],
+                                iopub_thread=self.iopub_thread,
                                 iopub_socket=self.iopub_socket,
                                 stdin_socket=self.stdin_socket,
                                 log=self.log,
                                 profile_dir=self.profile_dir,
                                 user_ns=self.user_ns,
         )
-        kernel.record_ports(self.ports)
+        kernel.record_ports({
+            name + '_port': port for name, port in self.ports.items()
+        })
         self.kernel = kernel
 
+        # Allow the displayhook to get the execution count
+        self.displayhook.get_execution_count = lambda: kernel.execution_count
+
     def init_gui_pylab(self):
         """Enable GUI event loop integration, taking pylab into account."""
 
+        # Register inline backend as default
+        # this is higher priority than matplotlibrc,
+        # but lower priority than anything else (mpl.use() for instance).
+        # This only affects matplotlib >= 1.5
+        if not os.environ.get('MPLBACKEND'):
+            os.environ['MPLBACKEND'] = 'module://ipykernel.pylab.backend_inline'
+
         # Provide a wrapper for :meth:`InteractiveShellApp.init_gui_pylab`
         # to ensure that any exception is printed straight to stderr.
         # Normally _showtraceback associates the reply with an execution,
@@ -331,8 +397,8 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
             # replace error-sending traceback with stderr
             def print_tb(etype, evalue, stb):
                 print ("GUI event loop or pylab initialization failed",
-                       file=io.stderr)
-                print (shell.InteractiveTB.stb2text(stb), file=io.stderr)
+                       file=sys.stderr)
+                print (shell.InteractiveTB.stb2text(stb), file=sys.stderr)
             shell._showtraceback = print_tb
             InteractiveShellApp.init_gui_pylab(self)
         finally:
@@ -342,7 +408,7 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         self.shell = getattr(self.kernel, 'shell', None)
         if self.shell:
             self.shell.configurables.append(self)
-    
+
     def init_extensions(self):
         super(IPKernelApp, self).init_extensions()
         # BEGIN HARDCODED WIDGETS HACK
@@ -355,17 +421,37 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
                 self.log.debug('ipywidgets package not installed.  Widgets will not be available.')
         # END HARDCODED WIDGETS HACK
 
+    def configure_tornado_logger(self):
+        """ Configure the tornado logging.Logger.
+
+            Must set up the tornado logger or else tornado will call
+            basicConfig for the root logger which makes the root logger
+            go to the real sys.stderr instead of the capture streams.
+            This function mimics the setup of logging.basicConfig.
+        """
+        logger = logging.getLogger('tornado')
+        handler = logging.StreamHandler()
+        formatter = logging.Formatter(logging.BASIC_FORMAT)
+        handler.setFormatter(formatter)
+        logger.addHandler(handler)
+
     @catch_config_error
     def initialize(self, argv=None):
         super(IPKernelApp, self).initialize(argv)
+        if self.subapp is not None:
+            return
+        # register zmq IOLoop with tornado
+        zmq_ioloop.install()
         self.init_blackhole()
         self.init_connection_file()
         self.init_poller()
         self.init_sockets()
         self.init_heartbeat()
         # writing/displaying connection info must be *after* init_sockets/heartbeat
-        self.log_connection_info()
         self.write_connection_file()
+        # Log connection info after writing connection file, so that the connection
+        # file is definitely available at the time someone reads the log.
+        self.log_connection_info()
         self.init_io()
         self.init_signal()
         self.init_kernel()
@@ -382,11 +468,14 @@ class IPKernelApp(BaseIPythonApplication, InteractiveShellApp,
         sys.stderr.flush()
 
     def start(self):
+        if self.subapp is not None:
+            return self.subapp.start()
         if self.poller is not None:
             self.poller.start()
         self.kernel.start()
+        self.io_loop = ioloop.IOLoop.current()
         try:
-            ioloop.IOLoop.instance().start()
+            self.io_loop.start()
         except KeyboardInterrupt:
             pass
 
